# beetcamp, Copyright (C) 2021 Šarūnas Nejus. Licensed under the GPLv2 or later.
# Based on beets-bandcamp: Copyright (C) 2015 Ariel George: Original implementation
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

"""Adds bandcamp album search support to the autotagger."""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import re
from functools import lru_cache, partial
from html import unescape
from operator import itemgetter, truth
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Union

import requests
from beets import IncludeLazyConfig, __version__, config, library, plugins
from beets.autotag.hooks import AlbumInfo, TrackInfo

from beetsplug import fetchart  # type: ignore[attr-defined]

from ._metaguru import Metaguru
from ._search import search_bandcamp

JSONDict = Dict[str, Any]

DEFAULT_CONFIG: JSONDict = {
    "include_digital_only_tracks": True,
    "search_max": 2,
    "art": False,
    "exclude_extra_fields": [],
    "genre": {
        "capitalize": False,
        "maximum": 0,
        "mode": "progressive",
        "always_include": [],
    },
    "comments_separator": "\n---\n",
}

ALBUM_URL_IN_TRACK = re.compile(r'<a id="buyAlbumLink" href="([^"]+)')
USER_AGENT = f"beets/{__version__} +http://beets.radbox.org/"


@lru_cache(maxsize=None)
def get_response(url: str) -> requests.Response:
    return requests.get(url, headers={"User-Agent": USER_AGENT})


class BandcampRequestsHandler:
    """A class that provides an ability to make requests and handles failures."""

    _log: logging.Logger
    config: IncludeLazyConfig

    def _exc(self, msg_template: str, *args: Sequence[str]) -> None:
        self._log.log(logging.WARNING, msg_template, *args, exc_info=True)

    def _info(self, msg_template: str, *args: Sequence[str]) -> None:
        self._log.log(logging.DEBUG, msg_template, *args, exc_info=False)

    def _get(self, url: str) -> str:
        """Return text contents of the url response."""
        response = get_response(url)
        try:
            response.raise_for_status()
        except requests.HTTPError:
            self._info("URL Not found: {}", url)
            return ""
        return unescape(response.text)

    def guru(self, url, attr):
        # type: (str, str) -> Optional[Union[TrackInfo, List[AlbumInfo]]]
        """Return Metaguru for the given URL."""
        kwargs = {}
        if hasattr(self, "config"):
            kwargs["config"] = self.config.flatten()
        try:
            return getattr(Metaguru.from_html(self._get(url), **kwargs), attr)
        except (KeyError, ValueError, AttributeError, IndexError):
            self._info("Failed obtaining {} from {}", attr, url)
        except Exception:  # pylint: disable=broad-except
            i_url = "https://github.com/snejus/beetcamp/issues/new"
            self._exc("Unexpected error obtaining {}, please report at {}", url, i_url)
        return None


def _from_bandcamp(clue: str) -> bool:
    """Check if the clue is likely to be a bandcamp url.
    We could check whether 'bandcamp' is found in the url, however, we would be ignoring
    cases where the publisher uses their own domain (for example https://eaux.ro) which
    in reality points to their Bandcamp page. Historically, we found that regardless
    of the domain, the rest of the url stays the same, therefore '/album/' or '/track/'
    is what we are looking for in a valid url here.
    """
    return bool(re.match(r"http[^ ]+/(album|track)/", clue))


class BandcampAlbumArt(BandcampRequestsHandler, fetchart.RemoteArtSource):
    NAME = "Bandcamp"

    def get(self, album, plugin, paths):
        # type: (AlbumInfo, plugins.BeetsPlugin, List[str]) -> Iterable[fetchart.Candidate]  # noqa
        """Return the url for the cover from the bandcamp album page.
        This only returns cover art urls for bandcamp albums (by id).
        """
        url = album.mb_albumid
        if not _from_bandcamp(url):
            self._info("Not fetching art for a non-bandcamp album URL")
        else:
            image = self.guru(url, "image")
            if image:
                yield self._candidate(url=image, match=fetchart.Candidate.MATCH_EXACT)


def urlify(pretty_string: str) -> str:
    """Transform a string into bandcamp url."""
    name = pretty_string.lower().replace("'", "").replace(".", "")
    return re.sub("--+", "-", re.sub(r"\W", "-", name, flags=re.ASCII)).strip("-")


class BandcampPlugin(BandcampRequestsHandler, plugins.BeetsPlugin):
    beets_config: IncludeLazyConfig

    def __init__(self) -> None:
        super().__init__()
        self.beets_config = config
        self.config.add(DEFAULT_CONFIG.copy())

        if self.config["art"]:
            self.register_listener("pluginload", self.loaded)

    @property
    def data_source(self) -> str:
        return "bandcamp"

    def loaded(self) -> None:
        """Add our own artsource to the fetchart plugin."""
        for plugin in plugins.find_plugins():
            if isinstance(plugin, fetchart.FetchArtPlugin):
                fetchart.ART_SOURCES[self.data_source] = BandcampAlbumArt
                fetchart.SOURCE_NAMES[BandcampAlbumArt] = self.data_source
                fetchart.SOURCES_ALL.append(self.data_source)
                bandcamp_fetchart = BandcampAlbumArt(self._log, self.config)
                plugin.sources = [bandcamp_fetchart, *plugin.sources]
                break

    def _find_url(self, item: library.Item, name: str, _type: str) -> str:
        """If the item has previously been imported, `mb_albumid` (or `mb_trackid`
        for singletons) contains the release url.

        As of 2022 April, Bandcamp purchases (at least in FLAC format) contain string
        *Visit {label_url}* in the `comments` field, therefore we try our luck here.

        If it is found, then the album/track name is converted into a valid url
        representation and appended to the `label_url`. This ends up being the correct
        url except when:
            * album name has been updated on Bandcamp but the file contains the old one
            * album name does not contain a single ascii alphanumeric character
              - in reality, this becomes '--{num}' in the url, where `num` depends on
              the number of previous releases that also did not have any valid
              alphanums. Therefore, we cannot make a reliable guess here.
        """
        url = getattr(item, f"mb_{_type}id", "")
        if _from_bandcamp(url):
            self._info("Fetching the URL attached to the first item, {}", url)
            return url

        m = re.match(r"Visit (https:[\w/.-]+com)", item.comments)
        urlified_name = urlify(name)
        if m and urlified_name:
            label = m.expand(r"\1")
            url = "/".join([label, _type, urlified_name])
            self._info("Trying our guess {} before searching", url)
            return url
        return ""

    def candidates(self, items, artist, album, va_likely, extra_tags=None):
        # type: (List[library.Item], str, str, bool, Any) -> Iterable[AlbumInfo]
        """Return a sequence of AlbumInfo objects that match the
        album whose items are provided or are being searched.
        """
        label = ""
        if items and album == items[0].album and artist == items[0].albumartist:
            label = items[0].label
            url = self._find_url(items[0], album, "album")
            if url:
                initial_guess = self.get_album_info(url)
                if initial_guess:
                    yield from initial_guess
                    return

        if "various" in artist.lower():
            artist = ""

        search = {"query": album, "artist": artist, "label": label, "search_type": "a"}
        results = map(itemgetter("url"), self._search(search))
        for res in filter(truth, map(self.get_album_info, results)):
            yield from res or [None]

    def item_candidates(self, item, artist, title):
        # type: (library.Item, str, str) -> Iterable[TrackInfo]
        """Return a sequence of TrackInfo objects that match the provided item.
        If the track was downloaded directly from bandcamp, it should contain
        a comment saying 'Visit <label-url>' - we look at this first by converting
        title into the format that Bandcamp use.
        """
        url = self._find_url(item, title, "track")
        label = ""
        if item and title == item.title and artist == item.artist:
            label = item.label
            initial_guess = self.get_track_info(url) if url else None
            if initial_guess:
                yield initial_guess
                return

        search = {"query": title, "artist": artist, "label": label, "search_type": "t"}
        results = map(itemgetter("url"), self._search(search))
        yield from filter(truth, map(self.get_track_info, results))

    def album_for_id(self, album_id: str) -> Optional[AlbumInfo]:
        """Fetch an album by its bandcamp ID."""
        if not _from_bandcamp(album_id):
            self._info("Not a bandcamp URL, skipping")
            return None

        albums = self.get_album_info(album_id)
        if not albums:
            return None

        if len(albums) > 1:
            # get the preferred media
            preferred = self.beets_config["match"]["preferred"]["media"].get()
            pref_to_idx = dict(zip(preferred, range(len(preferred))))
            albums = sorted(albums, key=lambda x: pref_to_idx.get(x.media, 100))
        return albums[0]

    def track_for_id(self, track_id: str) -> Optional[TrackInfo]:
        """Fetch a track by its bandcamp ID."""
        if _from_bandcamp(track_id):
            return self.get_track_info(track_id)

        self._info("Not a bandcamp URL, skipping")
        return None

    def get_album_info(self, url: str) -> Optional[List[AlbumInfo]]:
        """Return an AlbumInfo object for a bandcamp album page.
        If track url is given by mistake, find and fetch the album url instead.
        """
        html = self._get(url)
        if html and "/track/" in url:
            m = ALBUM_URL_IN_TRACK.search(html)
            if m:
                url = re.sub(r"/track/.*", m.expand(r"\1"), url)
        return self.guru(url, "albums")

    def get_track_info(self, url: str) -> Optional[TrackInfo]:
        """Returns a TrackInfo object for a bandcamp track page."""
        return self.guru(url, "singleton")

    def _search(self, data: JSONDict) -> Iterable[JSONDict]:
        """Return a list of track/album URLs of type search_type matching the query."""
        msg = "Searching {}s for {} using {}"
        self._info(msg, data["search_type"], data["query"], str(data))
        results = search_bandcamp(**data, get=self._get)
        return results[: self.config["search_max"].as_number()]


def get_args(args: List[str]) -> Any:
    from argparse import Action, ArgumentParser

    if TYPE_CHECKING:
        from argparse import Namespace

    parser = ArgumentParser(
        description="""Get bandcamp release metadata from the given <release-url>
or perform bandcamp search with <query>. Anything that does not start with https://
will be assumed to be a query.

Search type flags: -a for albums, -l for labels and artists, -t for tracks.
By default, all types are searched.
"""
    )

    class UrlOrQueryAction(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            # type: (Any, Namespace, Any, Any) -> None
            if values:
                if values.startswith("https://"):
                    target = "release_url"
                else:
                    target = "query"
                    del namespace.release_url
                setattr(namespace, target, values)

    exclusive = parser.add_mutually_exclusive_group()
    exclusive.add_argument(
        "release_url",
        action=UrlOrQueryAction,
        nargs="?",
        help="Release URL, starting with https:// OR",
    )
    exclusive.add_argument(
        "query", action=UrlOrQueryAction, default="", nargs="?", help="Search query"
    )

    store_const = partial(
        parser.add_argument, dest="search_type", action="store_const", default=""
    )
    store_const("-a", "--album", const="a", help="Search albums")
    store_const("-l", "--label", const="b", help="Search labels and artists")
    store_const("-t", "--track", const="t", help="Search tracks")
    parser.add_argument(
        "-o",
        "--open",
        action="store",
        dest="index",
        type=int,
        help="Open search result indexed by INDEX in the browser",
    )
    if not args:
        parser.print_help()
        parser.exit()
    return parser.parse_args(args=args)


def main() -> None:
    import json
    import sys

    args = get_args(sys.argv[1:])

    search_vars = vars(args)
    index = search_vars.pop("index", None)
    if search_vars.get("query"):
        search_results = search_bandcamp(**search_vars)

        if index:
            try:
                url = search_results[index - 1]["url"]
            except IndexError as e:
                raise Exception("Specified index could not be found") from e

            import webbrowser

            print(f"Opening search result number {index}: {url}")
            webbrowser.open(url)
        else:
            print(json.dumps(search_results))
    else:
        pl = BandcampPlugin()
        pl._log.setLevel(10)
        url = args.release_url
        result = pl.get_album_info(args.release_url) or pl.get_track_info(url)
        if not result:
            raise AssertionError("Failed to find a release under the given url")

        print(json.dumps(result))


if __name__ == "__main__":
    main()



"""Module with a Helpers class that contains various static, independent functions."""
import itertools as it
import operator as op
import re
from functools import lru_cache, partial
from typing import Any, Dict, Iterable, List, NamedTuple, Pattern

from beets.autotag.hooks import AlbumInfo
from ordered_set import OrderedSet as ordset

from .genres_lookup import GENRES

JSONDict = Dict[str, Any]
DIGI_MEDIA = "Digital Media"
FORMAT_TO_MEDIA = {
    "VinylFormat": "Vinyl",
    "CDFormat": "CD",
    "CassetteFormat": "Cassette",
    "DigitalFormat": DIGI_MEDIA,
    "DVDFormat": "DVD",
    "USB Flash Drive": DIGI_MEDIA,
}


class MediaInfo(NamedTuple):
    album_id: str
    name: str
    title: str
    description: str


CATALOGNUM_CONSTRAINT = r"""(?<![]/@-])(\b
(?!\W|LC[ ]|VA[\d ]+|[EL]P\W|[^\n.]+[ ](?:20\d\d|VA[ \d]+)|(?i:vol|disc|number|rd-9))
{}
\b(?!["%]))"""
_cat_pat = CATALOGNUM_CONSTRAINT.format(
    r"""
(
      [A-Z][A-Z .]+\d{3}         # HANDS D300, CC ATOM 101
    | [A-Z-]{3,}\d+              # RIV4
    | [A-Z]{2,}[A-Z.$-]*\d{2,}   # HS11, USE202, HEY-101, LI$INGLE025
    | (?<!\w\W)[A-Z.]{2,}[ ]\d+  # OBS.CUR 9
    | [A-z]+-[A-z]+[ ]?\d+       # o-ton 119
    | \w+[A-z]0\d+               # 1ØPILLS018, fa036
    | [a-z]+(?:cd|lp|:)\d+       # ostgutlp45, reni:7
    | [A-z]+\d+-\d+              # P90-003
)
(?: # optionally followed by
      (?<=\d\d)-?[A-Z]+  # IBM001CD (needs at least two digits before the letter)
    | [.]\d+             # ISMVA002.1
)?
"""
)

LABEL_CATNUM = CATALOGNUM_CONSTRAINT.format(r"(?i:{}[ ]?[A-Z]*\d+[A-Z]*)")
CATNUM_PAT = {
    # preceded by some variation of 'Catalogue number:'
    "header": re.compile(r"^cat[\w .]+(?:number\b:?|:) ?(\w.+)$", re.I | re.M),
    # beginning or end of line
    "start_end": re.compile(rf"(^{_cat_pat}|{_cat_pat}$)", re.M | re.VERBOSE),
    # enclosed by parens or square brackets, but not ending with MIX
    "delimited": re.compile(rf"(?:[\[(])(?!.*MIX){_cat_pat}(?:[])]|$)", re.VERBOSE),
    # can possibly be followed up by a second catalogue number
    "anywhere": re.compile(rf"({_cat_pat}(\ /\ {_cat_pat})?)", re.VERBOSE),
}

PATTERNS: Dict[str, Pattern[str]] = {
    "split_artists": re.compile(r", - |, | (?:[x+/-]|//|vs|and)[.]? "),
    "meta": re.compile(r'.*"@id".*'),
    "ft": re.compile(
        r"""
        [ ]*                     # all preceding space
        ((?P<br>[\[(])|\b)       # bracket or word boundary
        (ft|feat|featuring)[. ]  # one of the three ft variations
        (
            # when it does not start with a bracket, do not allow " - " in it, otherwise
            # we may match full track name
            (?(br)|(?!.*[ ]-[ ].*))
            # anything but brackets or a slash, except for a slash preceded
            # by a non-space (can be part of artist or title)
            (?:[^]\[()/]|\S/)+
        )
        (?<!mix)\b    # does not end with "mix"
        (?(br)[]\)])  # if it started with a bracket, it must end with a closing bracket
        [ ]*          # trailing space
    """,
        re.I | re.VERBOSE,
    ),
    "track_alt": re.compile(
        r"^([A-J]{1,3}[12]?\.?\d|[AB]+(?=\W{2,}))(?:(?!-\w)[^\w(]|_)+", re.I + re.M
    ),
    "vinyl_name": re.compile(r"[1-5](?= ?(xLP|LP|x))|single|double|triple", re.I),
}
rm_strings = [
    "limited edition",
    r"^[EL]P( \d+)?",
    r"\((digital )?album\)",
    r"\(single\)",
    r"^v/?a\W*|va$",
    r"\Wvinyl\W|vinyl-only",
    "compiled by.*",
    r"[\[(]presented by.*",
    r"free download|\([^()]*free(?!.*mix)[^()]*\)",
    "(\W|\W )bonus( \w+)*",
    r"[+][\w ]+remix|\(with remixes\)",
]

_remix_pat = r"(?P<remix>((?P<remixer>[^])]+) )?\b((re)?mix|edit|bootleg)\b[^])]*)"
# fmt: off
CLEAN_PATTERNS = [
    (re.compile(rf"(([\[(])|(^| ))\*?({'|'.join(rm_strings)})(?(2)[])]|( |$))", re.I), ""),       # noqa
    (re.compile(r" -(\S)"), r" - \1"),                    # hi -bye          -> hi - bye
    (re.compile(r"(\S)- "), r"\1 - "),                    # hi- bye          -> hi - bye
    (re.compile(r"  +"), " "),                            # hi  bye          -> hi bye
    (re.compile(r"(- )?\( *"), "("),                      # hi - ( bye)      -> hi (bye)
    (re.compile(r" \)+|(\)+$)"), ")"),                    # hi (bye ))       -> hi (bye)
    (re.compile(r"- Reworked"), "(Reworked)"),            # bye - Reworked   -> bye (Reworked)    # noqa
    (re.compile(rf"(\({_remix_pat})$", re.I), r"\1)"),    # bye - (Some Mix  -> bye - (Some Mix)  # noqa
    (re.compile(rf"- *({_remix_pat})$", re.I), r"(\1)"),  # bye - Some Mix   -> bye (Some Mix)    # noqa
    (re.compile(r'(^|- )[“"]([^”"]+)[”"]( \(|$)'), r"\1\2\3"),   # "bye" -> bye; hi - "bye" -> hi - bye  # noqa
    (re.compile(r"\((?i:(the )?(remixes))\)"), r"\2"),    # Album (Remixes)  -> Album Remixes     # noqa
]
# fmt: on


class Helpers:
    @staticmethod
    def get_label(meta: JSONDict) -> str:
        try:
            item = meta["albumRelease"][0]["recordLabel"]
        except (KeyError, IndexError):
            item = meta["publisher"]
        return item.get("name") or ""

    @staticmethod
    def get_vinyl_count(name: str) -> int:
        conv = {"single": 1, "double": 2, "triple": 3}
        for m in PATTERNS["vinyl_name"].finditer(name):
            count = m.group()
            return int(count) if count.isdigit() else conv[count.lower()]
        return 1

    @staticmethod
    def split_artists(artists: Iterable[str]) -> List[str]:
        """Split artists taking into account delimiters such as ',', '+', 'x', 'X' etc.
        Note: featuring artists are removed since they are not main artists.
        """
        no_ft_artists = (PATTERNS["ft"].sub("", a) for a in artists)
        split = map(PATTERNS["split_artists"].split, ordset(no_ft_artists))
        split_artists = ordset(map(str.strip, it.chain(*split))) - {"", "more"}

        for artist in list(split_artists):
            # ' & ' or ' X ' may be part of single artist name, so we need to be careful
            # here. We check whether any of the split artists appears on their own and
            # only split then
            for char in "X&":
                subartists = artist.split(f" {char} ")
                if len(subartists) > 1 and any(s in split_artists for s in subartists):
                    split_artists.discard(artist)  # type: ignore[attr-defined]
                    split_artists.update(subartists)  # type: ignore[attr-defined]
        return list(split_artists)

    @staticmethod
    @lru_cache(maxsize=None)
    def parse_catalognum(
        album="", disctitle="", description="", label="", artistitles=""
    ):
        # type: (str, str, str, str, str) -> str
        """Try getting the catalog number looking at text from various fields."""
        cases = [
            (CATNUM_PAT["header"], description),
            (CATNUM_PAT["anywhere"], disctitle),
            (CATNUM_PAT["anywhere"], album),
            (CATNUM_PAT["start_end"], description),
            (CATNUM_PAT["anywhere"], description),
        ]
        if label:
            pat = re.compile(LABEL_CATNUM.format(re.escape(label)), re.VERBOSE)
            cases.append((pat, "\n".join((album, disctitle, description))))

        def find(pat: Pattern[str], string: str) -> str:
            """Return the match if it is
            * not found in any of the track artists or titles
            * made of the label name when it has a space and is shorter than 6 chars
            """
            for m in pat.finditer(string):
                catnum = m.group(1).strip()
                if catnum.lower() not in artistitles:
                    if " " in catnum:
                        first = catnum.split()[0].lower()
                        if len(catnum) <= 5 and first not in label.lower():
                            continue
                    return catnum
            return ""

        try:
            return next(filter(None, it.starmap(find, cases)))
        except StopIteration:
            return ""

    @staticmethod
    def clean_name(name: str) -> str:
        for pat, repl in CLEAN_PATTERNS:
            name = pat.sub(repl, name).strip()
        return name

    @staticmethod
    def get_genre(keywords, config, label):
        # type: (Iterable[str], JSONDict, str) -> Iterable[str]
        """Return a comma-delimited list of valid genres, using MB genres for reference.

        Initially, exclude keywords that are label names (unless they are valid MB genres)

        Verify each keyword's (potential genre) validity w.r.t. the configured `mode`:
          * classical: valid only if the _entire keyword_ matches a MB genre in the list
          * progressive: either above or if each of the words matches MB genre - since it
            is effectively a subgenre.
          * psychedelic: either one of the above or if the last word is a valid MB genre.
            This allows to be flexible regarding the variety of potential genres while
            keeping away from spammy ones.

        Once we have the list of keywords that make it through the mode filters,
        an additional filter is executed:
          * if a keyword is _part of another keyword_ (genre within a sub-genre),
            the more generic option gets excluded, for example,
            >>> get_genre(['house', 'garage house', 'glitch'], "classical")
            'garage house, glitch'
        """
        valid_mb_genre = partial(op.contains, GENRES)
        label_name = label.lower().replace(" ", "")

        def is_label_name(kw: str) -> bool:
            return kw.replace(" ", "") == label_name and not valid_mb_genre(kw)

        def is_included(kw: str) -> bool:
            return any(re.search(x, kw) for x in config["always_include"])

        def valid_for_mode(kw: str) -> bool:
            if config["mode"] == "classical":
                return valid_mb_genre(kw)

            words = map(str.strip, kw.split(" "))
            if config["mode"] == "progressive":
                return valid_mb_genre(kw) or all(map(valid_mb_genre, words))

            return valid_mb_genre(kw) or valid_mb_genre(list(words)[-1])

        unique_genres: ordset[str] = ordset()
        # expand badly delimited keywords
        split_kw = partial(re.split, r"[.] | #| - ")
        for kw in it.chain.from_iterable(map(split_kw, keywords)):
            # remove full stops and hashes and ensure the expected form of 'and'
            kw = re.sub("[.#]", "", str(kw)).replace("&", "and")
            if not is_label_name(kw) and (is_included(kw) or valid_for_mode(kw)):
                unique_genres.add(kw)

        def duplicate(genre: str) -> bool:
            """Return True if genre is contained within another genre or if,
            having removed spaces from every other, there is a duplicate found.
            It is done this way so that 'dark folk' is kept while 'darkfolk' is removed,
            and not the other way around.
            """
            others = unique_genres - {genre}
            others = others.union(x.replace(" ", "").replace("-", "") for x in others)  # type: ignore[attr-defined] # noqa
            return any(genre in x for x in others)

        return it.filterfalse(duplicate, unique_genres)

    @staticmethod
    def unpack_props(obj: JSONDict) -> JSONDict:
        """Add all 'additionalProperty'-ies to the parent dictionary."""
        for prop in obj.get("additionalProperty") or []:
            obj[prop["name"]] = prop["value"]
        return obj

    @staticmethod
    def get_media_formats(format_list: List[JSONDict]) -> List[MediaInfo]:
        """Return filtered Bandcamp media formats as a list of MediaInfo objects.
        Formats are filtered using the following fields,

        type_id item_type  type_name          musicReleaseFormat
                a          Digital            DigitalFormat   # digital album
                b          Digital            DigitalFormat   # discography
                t          Digital            DigitalFormat   # digital track
        0       p          Other
        1       p          Compact Disc (CD)  CDFormat
        2       p          Vinyl LP           VinylFormat
        3       p          Cassette           CassetteFormat
        4       p          DVD                DVDFormat
        5       p          USB Flash Drive
        10      p          Poster/Print
        11      p          T-Shirt/Apparel
        15      p          2 x Vinyl LP       VinylFormat
        16      p          7" Vinyl           VinylFormat
        17      p          Vinyl Box Set      VinylFormat
        18      p          Other Vinyl        VinylFormat
        19      p          T-Shirt/Shirt
        20      p          Sweater/Hoodie
        """

        def valid_format(obj: JSONDict) -> bool:
            return (
                {"name", "item_type"} < set(obj)
                # not a discography
                and obj["item_type"] != "b"
                # musicReleaseFormat format is given or it is a USB
                and ("musicReleaseFormat" in obj or obj["type_id"] == 5)
                # it is not a vinyl bundle
                and not (obj["item_type"] == "p" and "bundle" in obj["name"].lower())
            )

        formats = []
        for _format in filter(valid_format, map(Helpers.unpack_props, format_list)):
            formats.append(
                MediaInfo(
                    _format["@id"],
                    FORMAT_TO_MEDIA[
                        _format.get("musicReleaseFormat") or "DigitalFormat"
                    ],
                    _format["name"],
                    _format.get("description") or "",
                )
            )
        return formats

    @staticmethod
    def add_track_alts(album: AlbumInfo, comments: str) -> AlbumInfo:
        # using an ordered set here in case of duplicates
        track_alts = ordset(PATTERNS["track_alt"].findall(comments))

        @lru_cache(maxsize=None)
        def get_medium_total(medium: int) -> int:
            starts = {1: "AB", 2: "CD", 3: "EF", 4: "GH", 5: "IJ"}[medium]
            return len(re.findall(rf"^[{starts}]", "\n".join(track_alts), re.M))

        medium = 1
        medium_index = 1
        if len(track_alts) == len(album.tracks):
            for track, track_alt in zip(album.tracks, track_alts):
                track.track_alt = track_alt
                track.medium_index = medium_index
                track.medium = medium
                track.medium_total = get_medium_total(medium)
                if track.medium_index == track.medium_total:
                    medium += 1
                    medium_index = 1
                else:
                    medium_index += 1
        return album



"""Module with bandcamp search functionality."""
import re
from difflib import SequenceMatcher
from html import unescape
from operator import itemgetter
from typing import Any, Callable, Dict, List

import requests

JSONDict = Dict[str, Any]
SEARCH_URL = "https://bandcamp.com/search?q={}"


def _f(field: str) -> str:
    """Return pattern matching a string that does not start with '<' or space
    until the end of the line.
    """
    return rf"(?P<{field}>[^\s<][^\n]+)"


RELEASE_PATTERNS = [
    re.compile(r"itemtype..\n\s+" + _f("type")),
    re.compile(r"search_item_type=[^>]+>\n\s+" + _f("name")),
    re.compile(r"\n\s+genre: " + _f("genre")),
    re.compile(r"\n\s+from " + _f("album")),
    re.compile(r"\n\s+by " + _f("artist")),
    re.compile(r"\n\s+released " + _f("date")),
    re.compile(r"\n\s+(?P<tracks>\d+) tracks"),
    re.compile(r">https://bandcamp\.(?P<label>[^.<]+)\.[^<]+<"),
    re.compile(r">https://(?P<label>[^.]+)\.bandcamp\.[^<]+<"),
    re.compile(r">https://(?P<label>(?!bandcamp)[^/]+)\.[^<]+<"),
    re.compile(r">(?P<url>https://[^<]+)<"),
]


def to_ascii(string: str) -> str:
    """Lowercase and translate non-ascii chars to '?'."""
    return string.lower().encode("ascii", "replace").decode()


def get_similarity(query: str, result: str) -> float:
    """Return the similarity between two strings normalized to [0, 1].
    We take into account how well the result matches the query, e.g.
        query: "foo"
        result: "foo bar"
    Similarity is then:
        (2 * (len("foo") / len("foo")) + len("foo") / len("foo bar")) / 3
    2/3 of the result is how much of the query is found in the result,
    and 1/3 is a penalty for the non-matching part.
    """
    a, b = to_ascii(query), to_ascii(result)
    if not a or not b:
        return 0
    m = SequenceMatcher(a=a, b=b).find_longest_match(0, len(a), 0, len(b))
    return ((m.size / len(a)) * 2 + m.size / len(b)) / 3


def get_matches(text: str) -> JSONDict:
    """Reduce matches from all patterns into a single dictionary."""
    result: JSONDict = {}
    for pat in RELEASE_PATTERNS:
        m = pat.search(text)
        if m:
            result = {**m.groupdict(), **result}
    if "type" in result:
        result["type"] = result["type"].lower()
    if "date" in result:
        result["date"] = " ".join(reversed(result["date"].split()))
    return result


def parse_and_sort_results(html: str, **kwargs: str) -> List[JSONDict]:
    """Given the html string, parse metadata for each entity and sort them
    by the field/value pairs given in kwargs.
    """
    results: List[JSONDict] = []
    for block in html.split("searchresult data-search")[1:]:
        similarities = []
        res = get_matches(block)
        for field, query in kwargs.items():
            similarities.append(get_similarity(query, res.get(field, "")))

        res["similarity"] = round(sum(similarities) / len(similarities), 3)
        results.append(res)
    results = sorted(results, key=itemgetter("similarity"), reverse=True)
    return [{"index": i + 1, **r} for i, r in enumerate(results)]


def get_url(url: str) -> str:
    response = requests.get(url)
    response.raise_for_status()
    return unescape(response.text)


def search_bandcamp(
    query: str = "",
    search_type: str = "",
    get: Callable[[str], str] = get_url,
    **kwargs: Any,
) -> List[JSONDict]:
    """Return a list with item JSONs of type search_type matching the query.
    Bandcamp search may be unpredictable, therefore search results get sorted
    regarding their similarity to what's being queried.
    """
    url = SEARCH_URL.format(query)
    if search_type:
        url += "&item_type=" + search_type
    kwargs["name"] = query
    return parse_and_sort_results(get(url), **kwargs)



"""Module with tracks parsing functionality."""
import itertools as it
import operator as op
import re
import sys
from collections import Counter
from dataclasses import dataclass
from functools import reduce
from typing import Iterator, List, Optional, Set

from ordered_set import OrderedSet

from ._helpers import CATNUM_PAT, Helpers, JSONDict, _remix_pat
from .track import Track

if sys.version_info.minor > 7:
    from functools import cached_property  # pylint: disable=ungrouped-imports
else:
    from cached_property import cached_property  # type: ignore # pylint: disable=import-error # noqa

digiwords = r"""
    # must contain at least one of
    (\W*(bandcamp|digi(tal)?|exclusive|bonus|bns|unreleased))+
    # and may be followed by
    (\W(track|only|tune))*
    """
DIGI_ONLY_PATTERN = re.compile(
    rf"""
\s*  # all preceding space
(
      (^{digiwords}[.:\d\s]+\s)     # begins with 'Bonus.', 'Bonus 1.' or 'Bonus :'
 | [\[(]{digiwords}[\])]\W*         # delimited by brackets, '[Bonus]', '(Bonus) -'
 |   [*]{digiwords}[*]              # delimited by asterisks, '*Bonus*'
 |  ([ ]{digiwords}$)               # might not be delimited if at the end, '... Bonus'
)
\s*  # all succeeding space
    """,
    re.I | re.VERBOSE,
)
DELIMITER_PAT = re.compile(r" ([^\w&()+/[\] ]) ")
ELP_ALBUM_PAT = re.compile(r"[- ]*\[([^\]]+ [EL]P)\]+")  # Title [Some Album EP]
TITLE_IN_QUOTES = re.compile(r'^(.+[^ -])[ -]+"([^"]+)"$')
NUMBER_PREFIX = re.compile(r"(^|- )\d{2,}\W* ")


@dataclass
class Tracks(List[Track]):
    tracks: List[Track]

    def __iter__(self) -> Iterator[Track]:
        return iter(self.tracks)

    def __len__(self) -> int:
        return len(self.tracks)

    @classmethod
    def from_json(cls, meta: JSONDict) -> "Tracks":
        try:
            tracks = [{**t, **t["item"]} for t in meta["track"]["itemListElement"]]
        except (TypeError, KeyError):
            tracks = [meta]

        names = [i.get("name", "") for i in tracks]
        names = cls.split_quoted_titles(names)
        names = cls.remove_number_prefix(names)
        delim = cls.track_delimiter(names)
        for track, name in zip(tracks, names):
            track["name_parts"] = {"clean": name}
            track["delim"] = delim
        tracks = cls.common_name_parts(tracks, names)
        return cls([Track.from_json(t, delim, Helpers.get_label(meta)) for t in tracks])

    @cached_property
    def first(self) -> Track:
        return self.tracks[0]

    @staticmethod
    def split_quoted_titles(names: List[str]) -> List[str]:
        if len(names) > 1 and all(TITLE_IN_QUOTES.match(n) for n in names):
            return [TITLE_IN_QUOTES.sub(r"\1 - \2", n) for n in names]
        return names

    @staticmethod
    def remove_number_prefix(names: List[str]) -> List[str]:
        if len(names) > 1 and all(NUMBER_PREFIX.search(n) for n in names):
            return [NUMBER_PREFIX.sub(r"\1", n) for n in names]
        return names

    @staticmethod
    def common_name_parts(tracks, names):
        # type: (List[JSONDict], List[str]) -> List[JSONDict]
        """Parse track names for parts that require knowledge of the other names.

        1. Split each track name into words
        2. Find the list of words that are common to all tracks
           a. check the *first* and the *last* word for the catalog number
              1. If found, remove it from every track name
           b. check whether tracks start the same way. This indicates an album with
              one unique root title and a couple of its remixes. This is especially
              relevant when the remix is not delimited appropriately.
        Return the catalog number and the new list of names.
        """
        names_tokens = map(str.split, names)
        common_words = reduce(op.and_, [OrderedSet(x) for x in names_tokens])
        if not common_words:
            return tracks

        matches = (CATNUM_PAT["anywhere"].search(common_words[i]) for i in [0, -1])
        try:
            cat, word = next(((m.group(1), m.string) for m in matches if m))
        except StopIteration:
            pass
        else:
            for track in tracks:
                track["name_parts"].update(
                    catalognum=cat,
                    clean=track["name_parts"]["clean"].replace(word, "").strip(),
                )

        joined = " ".join(common_words)
        if joined in names:  # it is one of the track names (root title)
            for track in tracks:
                leftover = track["name_parts"]["clean"].replace(joined, "").lstrip()
                # looking for a remix without brackets
                if re.fullmatch(_remix_pat, leftover, re.I):
                    track["name_parts"]["clean"] = f"{joined} ({leftover})"

        return tracks

    @cached_property
    def raw_names(self) -> List[str]:
        return [j.name for j in self.tracks]

    @cached_property
    def original_artists(self) -> List[str]:
        """Return all unique unsplit (original) main track artists."""
        return list(dict.fromkeys(j.artist for j in self.tracks))

    @property
    def artists(self) -> List[str]:
        """Return all unique split main track artists.

        "Artist1 x Artist2" -> ["Artist1", "Artist2"]
        """
        return list(dict.fromkeys(it.chain(*(j.artists for j in self.tracks))))

    @property
    def remixers(self) -> List[str]:
        """Return all remix artists."""
        return [
            t.remix.remixer
            for t in self.tracks
            if t.remix and not t.remix.by_other_artist
        ]

    @property
    def other_artists(self) -> Set[str]:
        """Return all unique remix and featuring artists."""
        ft = [j.ft for j in self.tracks if j.ft]
        return set(it.chain(self.remixers, ft))

    @cached_property
    def all_artists(self) -> Set[str]:
        """Return all unique (1) track, (2) remix, (3) featuring artists."""
        return self.other_artists | set(self.original_artists)

    @cached_property
    def artistitles(self) -> str:
        """Returned artists and titles joined into one long string."""
        return " ".join(it.chain(self.raw_names, self.all_artists)).lower()

    @cached_property
    def single_catalognum(self) -> Optional[str]:
        """Return catalognum if every track contains the same one."""
        cats = [t.catalognum for t in self if t.catalognum]
        if len(cats) == len(self) and len(set(cats)) == 1:
            return cats[0]
        return None

    @cached_property
    def albums_in_titles(self) -> Set[str]:
        return {t.album for t in self if t.album}

    def adjust_artists(self, albumartist: str) -> None:
        """Handle some track artist edge cases.

        These checks require knowledge of the entire release, therefore cannot be
        performed within the context of a single track.

        * When artist name is mistaken for the track_alt
        * When artist and title are delimited by '-' without spaces
        * When artist and title are delimited by a UTF-8 dash equivalent
        * Defaulting to the album artist
        """
        track_alts = {t.track_alt for t in self.tracks if t.track_alt}
        artists = [t.artist for t in self.tracks if t.artist]

        for t in [track for track in self.tracks if not track.artist]:
            if t.track_alt and len(track_alts) == 1:  # only one track_alt
                # the only track that parsed a track alt - it's most likely a mistake
                # one artist was confused for a track alt, like 'B2', - reverse this
                t.artist, t.track_alt = t.track_alt, None
            elif len(artists) == len(self) - 1:  # only 1 missing artist
                # if this is a remix and the parsed title is part of the albumartist or
                # is one of the track artists, we made a mistake parsing the remix:
                #  it is most probably the edge case where the `main_title` is a
                #  legitimate artist and the track title is something like 'Hello Remix'
                if t.remix and (t.main_title in albumartist):
                    t.artist, t.title = t.main_title, t.remix.remix
                # this is the only artist that didn't get parsed - relax the rule
                # and try splitting with '-' without spaces
                split = t.title.split("-")
                if len(split) == 1:
                    # attempt to split by another ' ? ' where '?' may be some utf-8
                    # alternative of a dash
                    split = [s for s in DELIMITER_PAT.split(t.title) if len(s) > 1]
                if len(split) > 1:
                    t.artist, t.title = split
            if not t.artist:
                # default to the albumartist
                t.artist = albumartist

    @staticmethod
    def track_delimiter(names: List[str]) -> str:
        """Return the track parts delimiter that is in effect in the current release.
        In some (rare) situations track parts are delimited by a pipe character
        or some UTF-8 equivalent of a dash.

        This checks every track for the first character (see the regex for exclusions)
        that splits it. The character that splits the most and at least half of
        the tracks is the character we need.

        If no such character is found, or if we have just one track, return a dash '-'.
        """

        def get_delim(string: str) -> str:
            m = DELIMITER_PAT.search(string)
            return m.group(1) if m else "-"

        delim, count = Counter(map(get_delim, names)).most_common(1).pop()
        return delim if (len(names) == 1 or count > len(names) / 2) else "-"



"""Module for parsing bandcamp metadata."""
import itertools as it
import json
import operator as op
import re
import sys
from collections import Counter
from datetime import date, datetime
from functools import partial
from typing import Any, Dict, Iterable, List, Optional, Set
from unicodedata import normalize

from beets import __version__ as beets_version
from beets import config as beets_config
from beets.autotag.hooks import AlbumInfo, TrackInfo
from pycountry import countries, subdivisions

from ._helpers import PATTERNS, Helpers, MediaInfo
from ._tracks import Track, Tracks
from .album import AlbumName

if sys.version_info.minor > 7:
    from functools import cached_property  # pylint: disable=ungrouped-imports
else:
    from cached_property import cached_property  # type: ignore # pylint: disable=import-error # noqa

NEW_BEETS = int(beets_version.split(".")[1]) > 4

JSONDict = Dict[str, Any]

COUNTRY_OVERRIDES = {
    "Russia": "RU",  # pycountry: Russian Federation
    "The Netherlands": "NL",  # pycountry: Netherlands
    "UK": "GB",  # pycountry: Great Britain
    "D.C.": "US",
    "South Korea": "KR",  # pycountry: Korea, Republic of
}
DATA_SOURCE = "bandcamp"
WORLDWIDE = "XW"
DIGI_MEDIA = "Digital Media"
VA = "Various Artists"


class Metaguru(Helpers):
    _singleton = False
    va_name = VA
    media = MediaInfo("", "", "", "")

    meta: JSONDict
    config: JSONDict
    media_formats: List[MediaInfo]
    _tracks: Tracks
    _album_name: AlbumName

    def __init__(self, meta: JSONDict, config: Optional[JSONDict] = None) -> None:
        self.meta = meta
        self.media_formats = self.get_media_formats(
            (meta.get("inAlbum") or meta).get("albumRelease") or []
        )
        if self.media_formats:
            self.media = self.media_formats[0]
        self.config = config or {}
        self.va_name = beets_config["va_name"].as_str() or self.va_name
        self._tracks = Tracks.from_json(meta)
        self._album_name = AlbumName(
            meta, self.all_media_comments, self._tracks.albums_in_titles
        )

    @classmethod
    def from_html(cls, html: str, config: Optional[JSONDict] = None) -> "Metaguru":
        try:
            meta = re.search(PATTERNS["meta"], html.replace("\u200b", "")).group()  # type: ignore[union-attr]  # noqa
        except AttributeError as exc:
            raise AttributeError("Could not find release metadata JSON") from exc
        else:
            return cls(json.loads(meta), config)

    @cached_property
    def excluded_fields(self) -> Set[str]:
        return set(self.config.get("excluded_fields") or [])

    @property
    def comments(self) -> str:
        """Return release, media descriptions and credits separated by
        the configured separator string.
        """
        parts: List[str] = [self.meta.get("description") or ""]
        media_desc = self.media.description
        if media_desc and not media_desc.startswith("Includes high-quality"):
            parts.append(media_desc)

        parts.append(self.meta.get("creditText") or "")
        sep: str = self.config["comments_separator"]
        return sep.join(filter(op.truth, parts)).replace("\r", "")

    @cached_property
    def all_media_comments(self) -> str:
        return "\n".join([*[m.description for m in self.media_formats], self.comments])

    @cached_property
    def label(self) -> str:
        m = re.search(r"Label:([^/,\n]+)", self.all_media_comments)
        if m:
            return m.expand(r"\1").strip(" '\"")

        return self.get_label(self.meta)

    @cached_property
    def album_id(self) -> str:
        return self.meta.get("@id") or ""

    @cached_property
    def artist_id(self) -> str:
        try:
            return self.meta["byArtist"]["@id"]  # type: ignore [no-any-return]
        except KeyError:
            return self.meta["publisher"]["@id"]  # type: ignore [no-any-return]

    @cached_property
    def original_albumartist(self) -> str:
        m = re.search(r"Artists?:([^\n]+)", self.all_media_comments)
        aartist = m.group(1).strip() if m else self.meta["byArtist"]["name"]
        return re.sub(r" +// +", ", ", aartist)

    @cached_property
    def original_album(self) -> str:
        return self._album_name.original

    @cached_property
    def bandcamp_albumartist(self) -> str:
        """Return the official release albumartist.
        It is correct in half of the cases. In others, we usually find the label name.
        """
        aartist = self.original_albumartist
        if self.label == aartist:
            split = AlbumName.clean(self.original_album, [self.catalognum]).split(" - ")
            if len(split) > 1:
                aartist = split[0]

        aartists = Helpers.split_artists([aartist])
        if len(aartists) == 1:
            return aartist

        remixers_str = " ".join(self._tracks.other_artists).lower()

        def not_remixer(x: str) -> bool:
            splits = {x, *x.split(" & ")}
            return not any(y.lower() in remixers_str for y in splits)

        valid = list(filter(not_remixer, aartists))
        if len(valid) == len(aartists) and len(self._tracks.artists) <= 4:
            return aartist
        return ", ".join(valid)

    @cached_property
    def image(self) -> str:
        image = self.meta.get("image") or ""
        if isinstance(image, list) and isinstance(image[0], str):
            return image[0]
        return image

    @cached_property
    def release_date(self) -> Optional[date]:
        """Parse the datestring that takes the format like below and return date object.
        {"datePublished": "17 Jul 2020 00:00:00 GMT"}

        If the field is not found, return None.
        """
        rel = self.meta.get("datePublished") or self.meta.get("dateModified")
        if rel:
            return datetime.strptime(re.sub(r" \d{2}:.+", "", rel), "%d %b %Y").date()
        return rel

    @cached_property
    def albumstatus(self) -> str:
        reldate = self.release_date
        return "Official" if reldate and reldate <= date.today() else "Promotional"

    @property
    def disctitle(self) -> str:
        """Return medium's disc title if found."""
        return "" if self.media.name == DIGI_MEDIA else self.media.title

    @property
    def mediums(self) -> int:
        return self.get_vinyl_count(self.disctitle) if self.media.name == "Vinyl" else 1

    @cached_property
    def general_catalognum(self) -> str:
        """Find catalog number in the media-agnostic release metadata and cache it."""
        return self._tracks.single_catalognum or self.parse_catalognum(
            album=self.meta["name"],
            description=self.comments,
            label=self.label if not self._singleton else "",
            artistitles=self._tracks.artistitles,
        )

    @property
    def catalognum(self) -> str:
        """Find catalog number in the media-specific release metadata or return
        the cached media-agnostic one.
        """
        return (
            self.parse_catalognum(
                disctitle=self.disctitle,
                description=self.media.description,
                label=self.label if not self._singleton else "",
                artistitles=self._tracks.artistitles,
            )
            or self.general_catalognum
        )

    @cached_property
    def country(self) -> str:
        try:
            loc = self.meta["publisher"]["foundingLocation"]["name"].rpartition(", ")[
                -1
            ]
            name = normalize("NFKD", loc).encode("ascii", "ignore").decode()
            return (
                COUNTRY_OVERRIDES.get(name)
                or getattr(countries.get(name=name, default=object), "alpha_2", None)
                or subdivisions.lookup(name).country_code
            )
        except (ValueError, LookupError):
            return WORLDWIDE

    @cached_property
    def tracks(self) -> Tracks:
        self._tracks.adjust_artists(self.bandcamp_albumartist)
        return self._tracks

    @cached_property
    def unique_artists(self) -> List[str]:
        return self.split_artists(self._tracks.artists)

    @cached_property
    def albumartist(self) -> str:
        """Take into account the release contents and return the actual albumartist.
        * 'Various Artists' (or `va_name` configuration option) for a compilation release
        """
        if self.va:
            return self.va_name

        if len(self._tracks) == 1:
            return self.tracks.first.artist

        aartist = self.original_albumartist
        if self.unique_artists:
            aartist = ", ".join(sorted(self.unique_artists))

        return aartist

    @cached_property
    def vinyl_disctitles(self) -> str:
        return " ".join([m.title for m in self.media_formats if m.name == "Vinyl"])

    @cached_property
    def album_name(self) -> str:
        return self._album_name.get(
            self.catalognum,
            self.tracks.original_artists,
            self.tracks.artists,
            self.label,
        )

    def _search_albumtype(self, word: str) -> bool:
        """Return whether the given word (ep or lp) matches the release albumtype.
        True when one of the following conditions is met:
        * if {word}[0-9] is found in the catalognum
        * if it's found in the original album name or any vinyl disctitle
        * if it's found in the same sentence as 'this' or '{album_name}', where
        sentences are read from release and media descriptions.
        """
        sentences = re.split(r"[.]\s+|\n", self.all_media_comments)
        word_pat = re.compile(rf"\b{word}\b", re.I)
        catnum_pat = re.compile(rf"{word}\d", re.I)
        name_pat = re.compile(rf"\b(this|{re.escape(self.album_name)})\b", re.I)
        return bool(
            catnum_pat.search(self.catalognum)
            or word_pat.search(self.original_album + " " + self.vinyl_disctitles)
            or any(word_pat.search(s) and name_pat.search(s) for s in sentences)
        )

    @cached_property
    def is_single_album(self) -> bool:
        return (
            self._singleton
            or len({t.main_title for t in self.tracks}) == 1
            or len(self._tracks.raw_names) == 1
        )

    @cached_property
    def is_lp(self) -> bool:
        """Return whether the release is an LP."""
        return self._search_albumtype("lp")

    @cached_property
    def is_ep(self) -> bool:
        """Return whether the release is an EP."""
        return self._search_albumtype("ep") or (
            " / " in self.album_name and len(self.tracks.artists) == 2
        )

    def check_albumtype_in_descriptions(self) -> str:
        """Count 'lp', 'album' and 'ep' words in the release and media descriptions
        and return the albumtype that represents the word matching the most times.
        """
        matches = re.findall(r"\b(album|ep|lp)\b", self.all_media_comments.lower())
        if matches:
            counts = Counter(x.replace("lp", "album") for x in matches)
            # if equal, we assume it's an EP since it's more likely that an EP is
            # referred to as an "album" rather than the other way around
            if counts["ep"] >= counts["album"]:
                return "ep"
        return "album"

    @cached_property
    def is_comp(self) -> bool:
        """Return whether the release is a compilation."""

        def first_one(artist: str) -> str:
            return PATTERNS["split_artists"].split(artist.replace(" & ", ", "))[0]

        truly_unique = set(map(first_one, self.tracks.artists))
        return (
            self._album_name.mentions_compilation
            or self._search_albumtype("compilation")
            or (len(truly_unique) > 3 and len(self.tracks) > 4)
        )

    @cached_property
    def albumtype(self) -> str:
        if self._singleton:
            return "single"
        if self.is_ep:
            return "ep"
        if self.is_lp:
            return "album"

        atype = self.check_albumtype_in_descriptions()
        if atype == "ep":
            return "ep"
        # otherwise, it's an album, but we firstly need to check if it's a compilation
        if self.is_comp:
            return "compilation"

        return "album"

    @cached_property
    def albumtypes(self) -> str:
        albumtypes = {self.albumtype}
        if self.is_comp:
            if self.albumtype == "ep":
                albumtypes.add("compilation")
            else:
                albumtypes.add("album")
        if self.is_lp:
            albumtypes.add("lp")
        if self.is_single_album:
            albumtypes.add("single")
        for word in ["remix", "rmx", "edits", "live", "soundtrack"]:
            if word in self.original_album.lower():
                albumtypes.add(word.replace("rmx", "remix").replace("edits", "remix"))
        if len(self.tracks.remixers) == len(self.tracks):
            albumtypes.add("remix")

        return "; ".join(sorted(albumtypes))

    @cached_property
    def va(self) -> bool:
        return len(self.unique_artists) > 3

    @cached_property
    def style(self) -> Optional[str]:
        """Extract bandcamp genre tag from the metadata."""
        # expecting the following form: https://bandcamp.com/tag/folk
        tag_url = self.meta.get("publisher", {}).get("genre") or ""
        style = None
        if tag_url:
            style = tag_url.split("/")[-1]
            if self.config["genre"]["capitalize"]:
                style = style.capitalize()
        return style

    @cached_property
    def genre(self) -> Optional[str]:
        kws: Iterable[str] = map(str.lower, self.meta["keywords"])
        if self.style:
            exclude_style = partial(op.ne, self.style.lower())
            kws = filter(exclude_style, kws)

        genre_cfg = self.config["genre"]
        genres = self.get_genre(kws, genre_cfg, self.label)
        if genre_cfg["capitalize"]:
            genres = map(str.capitalize, genres)
        if genre_cfg["maximum"]:
            genres = it.islice(genres, genre_cfg["maximum"])

        return ", ".join(sorted(genres)).strip() or None

    @property
    def _common(self) -> JSONDict:
        return {
            "data_source": DATA_SOURCE,
            "media": self.media.name,
            "data_url": self.album_id,
            "artist_id": self.artist_id,
        }

    def get_fields(self, fields: Iterable[str], src: object = None) -> JSONDict:
        """Return a mapping between unexcluded fields and their values."""
        fields = list(set(fields) - self.excluded_fields)
        if len(fields) == 1:
            field = fields.pop()
            return {field: getattr(self, field)}
        return dict(zip(fields, iter(op.attrgetter(*fields)(src or self))))

    @property
    def _common_album(self) -> JSONDict:
        common_data: JSONDict = {"album": self.album_name}
        fields = ["label", "catalognum", "albumtype", "country"]
        if NEW_BEETS:
            fields.extend(["genre", "style", "comments", "albumtypes"])
        common_data.update(self.get_fields(fields))
        reldate = self.release_date
        if reldate:
            common_data.update(self.get_fields(["year", "month", "day"], reldate))

        return common_data

    def _trackinfo(self, track: Track, **kwargs: Any) -> TrackInfo:
        data = track.info
        data.update(**self._common, **kwargs)
        # if track-level catalognum is not found or if it is the same as album's, then
        # remove it. Otherwise, keep it attached to the track
        if not data["catalognum"] or data["catalognum"] == self.catalognum:
            data.pop("catalognum", None)
        if not data["lyrics"]:
            data.pop("lyrics", None)
        if not NEW_BEETS:
            data.pop("catalognum", None)
            data.pop("lyrics", None)
        for field in set(data.keys()) & self.excluded_fields:
            data.pop(field)

        return TrackInfo(**data)

    @cached_property
    def singleton(self) -> TrackInfo:
        self._singleton = True
        self.media = self.media_formats[0]
        track = self._trackinfo(self.tracks.first)
        if NEW_BEETS:
            track.update(self._common_album)
            track.pop("album", None)
        track.track_id = track.data_url
        return track

    def get_media_album(self, media: MediaInfo) -> AlbumInfo:
        """Return album for the appropriate release format."""
        self.media = media
        include_digi = self.config.get("include_digital_only_tracks")

        tracks = list(self.tracks)
        if not include_digi and self.media.name != DIGI_MEDIA:
            tracks = [t for t in self.tracks if not t.digi_only]

        get_trackinfo = partial(
            self._trackinfo,
            medium=1,
            disctitle=self.disctitle or None,
            medium_total=len(self.tracks),
        )
        album_info = AlbumInfo(
            **self._common,
            **self._common_album,
            artist=self.albumartist,
            album_id=self.album_id,
            mediums=self.mediums,
            albumstatus=self.albumstatus,
            tracks=list(map(get_trackinfo, tracks)),
        )
        for key, val in self.get_fields(["va"]).items():
            setattr(album_info, key, val)
        album_info.album_id = self.media.album_id
        if self.media.name == "Vinyl":
            album_info = self.add_track_alts(album_info, self.comments)
        return album_info

    @cached_property
    def albums(self) -> Iterable[AlbumInfo]:
        """Return album for the appropriate release format."""
        return list(map(self.get_media_album, self.media_formats))



"""Module with album parsing logic."""
import sys
from dataclasses import dataclass
import re
from typing import List, Set, Dict, Any

from ._helpers import PATTERNS, Helpers

if sys.version_info.minor > 7:
    from functools import cached_property  # pylint: disable=ungrouped-imports
else:
    from cached_property import cached_property  # type: ignore # pylint: disable=import-error # noqa

JSONDict = Dict[str, Any]


@dataclass
class AlbumName:
    _series = r"(?i:\b(part|volume|pt|vol)\b\.?)"
    SERIES = re.compile(rf"{_series}[ ]?[A-Z\d.-]+\b")
    SERIES_FMT = re.compile(rf"^(.+){_series} *0*")
    INCL = re.compile(r"[^][\w]*inc[^()]+mix(es)?[^()-]*\W?", re.I)
    EPLP = re.compile(r"\S*(?:Double )?(\b[EL]P\b)\S*", re.I)

    meta: JSONDict
    description: str
    albums_in_titles: Set[str]

    remove_artists = True

    @cached_property
    def in_description(self) -> str:
        """Check description for the album name header and return whatever follows it
        if found.
        """
        m = re.search(r"(Title: ?|Album(:|/Single) )([^\n]+)", self.description)
        if m:
            self.remove_artists = False
            return m.group(3).strip()
        return ""

    @cached_property
    def original(self) -> str:
        return self.meta.get("name") or ""

    @cached_property
    def mentions_compilation(self) -> bool:
        return bool(re.search(r"compilation|best of|anniversary", self.original, re.I))

    @cached_property
    def parsed(self) -> str:
        """
        Search for the album name in the following order and return the first match:
        1. Album name is found in *all* track names
        2. When 'EP' or 'LP' is in the release name, album name is what precedes it.
        3. If some words are enclosed in quotes in the release name, it is assumed
           to be the album name. Remove the quotes in such case.
        """
        if len(self.albums_in_titles) == 1:
            return next(iter(self.albums_in_titles))

        album = self.original
        for pat in [
            r"(((&|#?\b(?!Double|VA|Various)(\w|[^\w| -])+) )+[EL]P)",
            r"((['\"])([^'\"]+)\2( VA\d+)*)( |$)",
        ]:
            m = re.search(pat, album)
            if m:
                album = m.group(1).strip()
                return re.sub(r"^['\"](.+)['\"]$", r"\1", album)
        return album

    @cached_property
    def album_sources(self) -> List[str]:
        return list(filter(None, [self.in_description, self.parsed, self.original]))

    @cached_property
    def name(self) -> str:
        return self.in_description or self.parsed or self.original

    @cached_property
    def series(self) -> str:
        m = self.SERIES.search("\n".join(self.album_sources))
        return m.group() if m else ""

    @staticmethod
    def format_series(m: re.Match) -> str:  # type: ignore[type-arg]
        """Format the series part in an album.

        * Ensure 'Vol' or 'Pt' is suffixed by '.'
        * Ensure that it is followed by a space
        * Leading zeroes from the number are already removed by 'SERIES_FMT'.
        """
        album, series = m.groups()
        if album[0].isupper() and not series[0].isupper():
            series = series.capitalize()

        suffix = "." if len(series) in {2, 3} else ""
        return f"{album}{series}{suffix} "

    def standardize_series(self, album: str) -> str:
        """Standardize 'Vol', 'Part' etc. format."""
        series = self.series
        if not series:
            return album

        if series.lower() not in album.lower():
            # series was not given in the description, but found in the original name
            if series[0].isalpha():
                series = f", {series}"

            album += series
        else:
            # move from the beginning to the end of the album
            album, moved = re.subn(rf"^({series})\W+(.+)", r"\2, \1", album)
            if not moved:
                # otherwise, ensure that it is delimited by a comma
                album = re.sub(rf"(?<=\w)( {series}(?!\)))", r",\1", album)

        return self.SERIES_FMT.sub(self.format_series, album)

    @staticmethod
    def remove_label(name: str, label: str) -> str:
        if not label:
            return name

        pattern = re.compile(
            rf"""
            \W*               # pick up any punctuation
            (?<!\w[ ])        # cannot be preceded by a simple word
            \b{re.escape(label)}\b
            (?![ -][A-Za-z])  # cannot be followed by a word
            ([^[\]\w]|\d)*    # pick up any digits and punctuation
        """,
            flags=re.VERBOSE | re.IGNORECASE,
        )
        return pattern.sub(" ", name).strip()

    @classmethod
    def clean(cls, name: str, to_clean: List[str], label: str = "") -> str:
        """Return clean album name.

        Catalogue number and artists to be removed are provided as 'to_clean'.
        """
        name = PATTERNS["ft"].sub(" ", name)
        name = re.sub(r"^\[(.*)\]$", r"\1", name)

        escaped = [re.escape(x) for x in filter(None, to_clean)] + [
            r"Various Artists?\b(?! [A-z])( \d+)?"
        ]
        for arg in escaped:
            name = re.sub(rf" *(?i:(compiled )?by|vs|\W*split w) {arg}", "", name)
            if not re.search(rf"\w {arg} \w|of {arg}", name, re.I):
                name = re.sub(
                    rf"(^|[^'\])\w]|_|\b)+(?i:{arg})([^'(\[\w]|_|(\d+$))*", " ", name
                ).strip()

        name = cls.remove_label(Helpers.clean_name(name), label)
        name = cls.INCL.sub("", name).strip("- ")

        # uppercase EP and LP, and remove surrounding parens / brackets
        name = cls.EPLP.sub(lambda x: x.group(1).upper(), name)
        return name.strip(" /")

    def check_eplp(self, album: str) -> str:
        """Return album name followed by 'EP' or 'LP' if that's given in the comments.

        When album is given, search for the album.
        Otherwise, search for (Capital-case Album Name) (EP or LP) and return the match.
        """
        if album:
            look_for = re.escape(f"{album} ")
        else:
            look_for = r"((?!The|This)\b[A-Z][^ \n]+\b )+"

        m = re.search(rf"{look_for}[EL]P", self.description)
        return m.group() if m else album

    def get(
        self,
        catalognum: str,
        original_artists: List[str],
        artists: List[str],
        label: str,
    ) -> str:
        album = self.name
        to_clean = [catalognum]
        if self.remove_artists:
            to_clean.extend(original_artists + artists)

        album = self.clean(album, sorted(to_clean, key=len, reverse=True), label)
        if album.startswith("("):
            album = self.name

        album = self.check_eplp(self.standardize_series(album))

        if "split ep" in album.lower() or (not album and len(artists) == 2):
            album = " / ".join(artists)

        return album or catalognum or self.name



"""List of MusicBrainz genres from https://beta.musicbrainz.org/genres"""

GENRES = {
    "2 tone",
    "2-step",
    "acid house",
    "acid jazz",
    "acid rock",
    "acid techno",
    "acid trance",
    "acoustic blues",
    "acoustic rock",
    "afoxê",
    "african blues",
    "afrobeat",
    "aggrotech",
    "alternative country",
    "alternative dance",
    "alternative folk",
    "alternative hip hop",
    "alternative metal",
    "alternative pop",
    "alternative punk",
    "alternative rock",
    "ambient",
    "ambient dub",
    "ambient house",
    "ambient pop",
    "ambient techno",
    "ambient trance",
    "american primitive guitar",
    "americana",
    "anarcho-punk",
    "anatolian rock",
    "andalusian classical",
    "anti-folk",
    "aor",
    "appalachian folk",
    "arena rock",
    "ars antiqua",
    "ars nova",
    "ars subtilior",
    "art pop",
    "art punk",
    "art rock",
    "atmospheric black metal",
    "avant-garde",
    "avant-garde jazz",
    "avant-garde metal",
    "avant-garde pop",
    "avant-prog",
    "axé",
    "bachata",
    "baião",
    "balearic beat",
    "ballad",
    "baltimore club",
    "barbershop",
    "bardcore",
    "baroque",
    "baroque pop",
    "bass house",
    "bassline",
    "battle rap",
    "batucada",
    "beat music",
    "bebop",
    "benga",
    "berlin school",
    "bhangra",
    "big band",
    "big beat",
    "black metal",
    "blackened death metal",
    "blackgaze",
    "bleep techno",
    "blue-eyed soul",
    "bluegrass",
    "bluegrass gospel",
    "blues",
    "blues rock",
    "bolero",
    "bolero son",
    "bongo flava",
    "boogie rock",
    "boogie-woogie",
    "boom bap",
    "bossa nova",
    "bounce",
    "breakbeat",
    "breakbeat hardcore",
    "breakcore",
    "breaks",
    "brega",
    "brega funk",
    "brill building",
    "brit funk",
    "britcore",
    "britpop",
    "broken beat",
    "brostep",
    "brutal death metal",
    "bubblegum bass",
    "bubblegum pop",
    "cabaret",
    "cajun",
    "calypso",
    "canción melódica",
    "candombe",
    "canterbury scene",
    "cantopop",
    "canzone napoletana",
    "cape jazz",
    "carimbó",
    "carnatic classical",
    "celtic",
    "celtic punk",
    "chacarera",
    "chachachá",
    "chalga",
    "chamamé",
    "chamber pop",
    "champeta",
    "changüí",
    "chanson à texte",
    "chanson française",
    "chicago blues",
    "chicano rap",
    "chillout",
    "chillstep",
    "chillwave",
    "chimurenga",
    "chiptune",
    "chopped and screwed",
    "choro",
    "christian metal",
    "christian rock",
    "christmas music",
    "chutney",
    "city pop",
    "classic blues",
    "classic country",
    "classic jazz",
    "classic rock",
    "classical",
    "classical crossover",
    "cloud rap",
    "club",
    "coco",
    "coldwave",
    "comedy",
    "comedy hip hop",
    "comedy rock",
    "compas",
    "complextro",
    "conscious hip hop",
    "contemporary christian",
    "contemporary classical",
    "contemporary folk",
    "contemporary gospel",
    "contemporary jazz",
    "contemporary r&b",
    "contra",
    "cool jazz",
    "copla",
    "corrido",
    "country",
    "country blues",
    "country boogie",
    "country folk",
    "country pop",
    "country rap",
    "country rock",
    "country soul",
    "coupé-décalé",
    "cowpunk",
    "crossover jazz",
    "crossover prog",
    "crossover thrash",
    "crunk",
    "crust punk",
    "cumbia",
    "cumbia villera",
    "cuplé",
    "cyber metal",
    "cybergrind",
    "cyberpunk",
    "d-beat",
    "dance",
    "dance-pop",
    "dance-punk",
    "dance-rock",
    "dancehall",
    "dangdut",
    "dansband",
    "danzón",
    "dark ambient",
    "dark cabaret",
    "dark electro",
    "dark folk",
    "dark jazz",
    "dark psytrance",
    "dark wave",
    "darkstep",
    "darksynth",
    "death 'n' roll",
    "death industrial",
    "death metal",
    "death-doom metal",
    "deathcore",
    "deathgrind",
    "deathrock",
    "deathstep",
    "deconstructed club",
    "deep funk",
    "deep house",
    "delta blues",
    "denpa",
    "depressive black metal",
    "descarga",
    "desert blues",
    "desert rock",
    "detroit techno",
    "dhrupad",
    "digicore",
    "digital hardcore",
    "disco",
    "disco polo",
    "dixieland",
    "djent",
    "doo-wop",
    "doom metal",
    "doomcore",
    "downtempo",
    "dream pop",
    "dreampunk",
    "drill",
    "drill and bass",
    "drone",
    "drone metal",
    "drum and bass",
    "dub",
    "dub poetry",
    "dub techno",
    "dubstep",
    "dungeon synth",
    "east coast hip hop",
    "easy listening",
    "easycore",
    "ebm",
    "edm",
    "electric blues",
    "electro",
    "electro house",
    "electro swing",
    "electro-funk",
    "electro-industrial",
    "electroclash",
    "electronic",
    "electronic rock",
    "electronica",
    "electronicore",
    "electropop",
    "electropunk",
    "electrotango",
    "emo",
    "emo pop",
    "emo rap",
    "emocore",
    "enka",
    "ethereal wave",
    "euro house",
    "euro-disco",
    "eurobeat",
    "eurodance",
    "europop",
    "exotica",
    "experimental",
    "experimental hip hop",
    "experimental rock",
    "extratone",
    "fado",
    "fidget house",
    "filk",
    "finnish tango",
    "flamenco",
    "flamenco jazz",
    "folk",
    "folk metal",
    "folk pop",
    "folk punk",
    "folk rock",
    "folktronica",
    "footwork",
    "forró",
    "forró eletrônico",
    "forró universitário",
    "freak folk",
    "freakbeat",
    "free folk",
    "free improvisation",
    "free jazz",
    "french house",
    "frevo",
    "fuji",
    "funaná",
    "funeral doom metal",
    "funk",
    "funk carioca",
    "funk metal",
    "funk rock",
    "funk soul",
    "funkot",
    "funktronica",
    "funky house",
    "fusion",
    "future bass",
    "future funk",
    "future garage",
    "future house",
    "future jazz",
    "futurepop",
    "g-funk",
    "gabber",
    "gagaku",
    "gangsta rap",
    "garage house",
    "garage punk",
    "garage rock",
    "ghazal",
    "ghetto house",
    "ghettotech",
    "glam",
    "glam metal",
    "glam punk",
    "glam rock",
    "glitch",
    "glitch hop",
    "glitch pop",
    "go-go",
    "goa trance",
    "goregrind",
    "gospel",
    "gothic",
    "gothic country",
    "gothic metal",
    "gothic rock",
    "grebo",
    "gregorian chant",
    "grime",
    "grindcore",
    "groove metal",
    "group sounds",
    "grunge",
    "guaguancó",
    "guajira",
    "guaracha",
    "gypsy jazz",
    "gypsy punk",
    "happy hardcore",
    "hard bop",
    "hard house",
    "hard rock",
    "hard trance",
    "hardbass",
    "hardcore hip hop",
    "hardcore punk",
    "hardcore techno",
    "hardstyle",
    "hardvapour",
    "harsh noise",
    "harsh noise wall",
    "hauntology",
    "heartland rock",
    "heavy metal",
    "heavy psych",
    "heavy rock",
    "hi-nrg",
    "highlife",
    "hindustani classical",
    "hip hop",
    "hip house",
    "hiplife",
    "honky tonk",
    "hopepunk",
    "horror punk",
    "horrorcore",
    "house",
    "hyperpop",
    "hyphy",
    "idm",
    "illbient",
    "indian pop",
    "indie",
    "indie folk",
    "indie pop",
    "indie rock",
    "indietronica",
    "indorock",
    "industrial",
    "industrial hardcore",
    "industrial hip hop",
    "industrial metal",
    "industrial musical",
    "industrial rock",
    "industrial techno",
    "instrumental",
    "instrumental jazz",
    "instrumental rock",
    "irish folk",
    "italo dance",
    "italo-disco",
    "j-core",
    "j-pop",
    "j-rock",
    "jaipongan",
    "jazz",
    "jazz blues",
    "jazz fusion",
    "jazz rap",
    "jazz rock",
    "jazz-funk",
    "joik",
    "jùjú",
    "juke",
    "jump blues",
    "jump up",
    "jungle",
    "jungle terror",
    "k-pop",
    "kabarett",
    "kaseko",
    "kasékò",
    "kawaii future bass",
    "kawaii metal",
    "kayōkyoku",
    "keroncong",
    "khyal",
    "kizomba",
    "kleinkunst",
    "klezmer",
    "krautrock",
    "kuduro",
    "kwaito",
    "latin",
    "latin ballad",
    "latin jazz",
    "latin pop",
    "latin rock",
    "latin soul",
    "leftfield",
    "levenslied",
    "line dance",
    "liquid funk",
    "lo-fi",
    "lo-fi hip hop",
    "lolicore",
    "louisiana blues",
    "lounge",
    "lovers rock",
    "lowercase",
    "luk krung",
    "luk thung",
    "madchester",
    "mainstream rock",
    "makossa",
    "maloya",
    "mambo",
    "mandopop",
    "mangue beat",
    "manila sound",
    "march",
    "marchinha",
    "mariachi",
    "marrabenta",
    "martial industrial",
    "maskanda",
    "math rock",
    "mathcore",
    "mbalax",
    "mbaqanga",
    "mbube",
    "medieval",
    "melodic black metal",
    "melodic death metal",
    "melodic dubstep",
    "melodic hardcore",
    "melodic metalcore",
    "melodic rock",
    "melodic trance",
    "mento",
    "merengue",
    "metal",
    "metalcore",
    "miami bass",
    "microhouse",
    "microsound",
    "milonga",
    "min'yō",
    "mincecore",
    "minimal",
    "minimal drum and bass",
    "minimal techno",
    "minimal wave",
    "mod",
    "modal jazz",
    "modern blues",
    "modern classical",
    "modern country",
    "mood kayō",
    "moombahcore",
    "moombahton",
    "mor lam",
    "morna",
    "motown",
    "mpb",
    "musical",
    "musique concrète",
    "nashville sound",
    "nature sounds",
    "nederpop",
    "neo soul",
    "neo-progressive rock",
    "neo-psychedelia",
    "neo-rockabilly",
    "neo-traditional country",
    "neoclassical dark wave",
    "neoclassical metal",
    "neoclassicism",
    "neofolk",
    "nerdcore",
    "neue deutsche härte",
    "neue deutsche welle",
    "neurofunk",
    "neurohop",
    "new age",
    "new beat",
    "new jack swing",
    "new romantic",
    "new wave",
    "nightcore",
    "no wave",
    "noise",
    "noise pop",
    "noise rock",
    "noisecore",
    "non-music",
    "norteño",
    "northern soul",
    "nova cançó",
    "nu disco",
    "nu jazz",
    "nu metal",
    "nu skool breaks",
    "nueva canción",
    "nueva trova",
    "nuevo flamenco",
    "nuevo tango",
    "nyū myūjikku",
    "occult rock",
    "oi",
    "old school death metal",
    "old-time",
    "onkyo",
    "opera",
    "operatic pop",
    "orchestral",
    "orchestral jazz",
    "outlaw country",
    "p-funk",
    "pachanga",
    "pagodão",
    "pagode",
    "pagode romântico",
    "paisley underground",
    "palm-wine",
    "parang",
    "partido alto",
    "pasodoble",
    "philly soul",
    "phleng phuea chiwit",
    "phonk",
    "piedmont blues",
    "plena",
    "plunderphonics",
    "political hip hop",
    "polka",
    "pop",
    "pop metal",
    "pop punk",
    "pop rap",
    "pop rock",
    "pop soul",
    "pornogrind",
    "porro",
    "post-bop",
    "post-classical",
    "post-grunge",
    "post-hardcore",
    "post-metal",
    "post-minimalism",
    "post-punk",
    "post-rock",
    "power electronics",
    "power metal",
    "power noise",
    "power pop",
    "powerviolence",
    "production music",
    "progressive",
    "progressive bluegrass",
    "progressive country",
    "progressive folk",
    "progressive house",
    "progressive metal",
    "progressive pop",
    "progressive rock",
    "progressive trance",
    "psybient",
    "psychedelic",
    "psychedelic folk",
    "psychedelic pop",
    "psychedelic rock",
    "psychobilly",
    "psytrance",
    "pub rock",
    "punk",
    "punk blues",
    "punk rap",
    "punk rock",
    "qawwali",
    "queercore",
    "quiet storm",
    "r&b",
    "ragga",
    "ragga hip-hop",
    "ragga jungle",
    "ragtime",
    "raï",
    "ranchera",
    "rap metal",
    "rap rock",
    "rapcore",
    "rautalanka",
    "rave",
    "rebetiko",
    "red dirt",
    "red song",
    "reggae",
    "reggaeton",
    "rhythmic noise",
    "ritual ambient",
    "rock",
    "rock and roll",
    "rock opera",
    "rockabilly",
    "rocksteady",
    "romantic classical",
    "roots reggae",
    "roots rock",
    "rumba",
    "runo song",
    "ryūkōka",
    "salegy",
    "salsa",
    "samba",
    "samba-canção",
    "samba-choro",
    "samba-exaltação",
    "samba-reggae",
    "samba-rock",
    "sasscore",
    "schlager",
    "schranz",
    "screamo",
    "sea shanty",
    "séga",
    "sertanejo",
    "sertanejo raiz",
    "sertanejo romântico",
    "sertanejo universitário",
    "shibuya-kei",
    "shoegaze",
    "sierreño",
    "singer-songwriter",
    "ska",
    "ska punk",
    "skacore",
    "skate punk",
    "skiffle",
    "skweee",
    "slam death metal",
    "slow waltz",
    "slowcore",
    "sludge metal",
    "smooth jazz",
    "smooth soul",
    "soca",
    "soft rock",
    "son cubano",
    "son montuno",
    "songo",
    "sophisti-pop",
    "soukous",
    "soul",
    "soul blues",
    "soul jazz",
    "southern gospel",
    "southern metal",
    "southern rock",
    "southern soul",
    "space age pop",
    "space ambient",
    "space disco",
    "space rock",
    "spectralism",
    "speed garage",
    "speed metal",
    "speedcore",
    "spoken word",
    "steampunk",
    "stoner metal",
    "stoner rock",
    "street punk",
    "stride",
    "surf rock",
    "swamp pop",
    "swamp rock",
    "swing",
    "symphonic black metal",
    "symphonic metal",
    "symphonic prog",
    "symphonic rock",
    "symphony",
    "synth funk",
    "synth-pop",
    "synthwave",
    "taarab",
    "tango",
    "tech house",
    "tech trance",
    "technical death metal",
    "techno",
    "techno bass",
    "techno kayō",
    "techstep",
    "tecnobrega",
    "teen pop",
    "tejano",
    "terrorcore",
    "texas blues",
    "thrash metal",
    "thrashcore",
    "thumri",
    "timba",
    "tizita",
    "traditional country",
    "trance",
    "trap",
    "trap edm",
    "tribal ambient",
    "tribal house",
    "trip hop",
    "tropical house",
    "tropicália",
    "trot",
    "trova",
    "tsapiky",
    "turbo-folk",
    "turntablism",
    "twee pop",
    "uk drill",
    "uk funky",
    "uk garage",
    "uk hardcore",
    "underground hip hop",
    "urban cowboy",
    "vallenato",
    "vaportrap",
    "vaporwave",
    "vaudeville",
    "viking metal",
    "visual kei",
    "vocal house",
    "vocal jazz",
    "vocal trance",
    "waltz",
    "wave",
    "west coast hip hop",
    "west coast swing",
    "western swing",
    "witch house",
    "wonky",
    "xote",
    "yacht rock",
    "yé-yé",
    "zamba",
    "zamrock",
    "zarzuela",
    "zeuhl",
    "zouk",
    "zydeco",
}



"""Module with a single track parsing functionality."""
import re
import sys
from dataclasses import dataclass, field
from typing import List, Optional, Tuple


from ._helpers import CATNUM_PAT, PATTERNS, Helpers, JSONDict, _remix_pat

if sys.version_info.minor > 7:
    from functools import cached_property  # pylint: disable=ungrouped-imports
else:
    from cached_property import cached_property  # type: ignore # pylint: disable=import-error # noqa

digiwords = r"""
    # must contain at least one of
    (\W*(bandcamp|digi(tal)?|exclusive|bonus|bns|unreleased))+
    # and may be followed by
    (\W(track|only|tune))*
    """
DIGI_ONLY_PATTERN = re.compile(
    rf"""
\s*  # all preceding space
(
      (^{digiwords}[.:\d\s]+\s)     # begins with 'Bonus.', 'Bonus 1.' or 'Bonus :'
 | [\[(]{digiwords}[\])]\W*         # delimited by brackets, '[Bonus]', '(Bonus) -'
 |   [*]{digiwords}[*]              # delimited by asterisks, '*Bonus*'
 |  ([ ]{digiwords}$)               # might not be delimited if at the end, '... Bonus'
)
\s*  # all succeeding space
    """,
    re.I | re.VERBOSE,
)
DELIMITER_PAT = re.compile(r" ([^\w&()+/[\] ]) ")
ELP_ALBUM_PAT = re.compile(r"[- ]*\[([^\]]+ [EL]P)\]+")  # Title [Some Album EP]
TITLE_IN_QUOTES = re.compile(r'^(.+[^ -])[ -]+"([^"]+)"$')
NUMBER_PREFIX = re.compile(r"(^|- )\d{2,}\W* ")


@dataclass
class Remix:
    PATTERN = re.compile(rf" *[\[(] *{_remix_pat}[])]", re.I)

    delimited: str
    remixer: str
    remix: str
    by_other_artist: bool

    @classmethod
    def from_name(cls, name: str) -> Optional["Remix"]:
        m = cls.PATTERN.search(name)
        if m:
            remix = m.groupdict()
            remix["delimited"] = m.group().strip()
            remix["remixer"] = remix["remixer"] or ""
            return cls(**remix, by_other_artist="Original" in remix["remix"])
        return None


@dataclass
class Track:
    json_item: JSONDict = field(default_factory=dict)
    track_id: str = ""
    index: Optional[int] = None
    json_artist: str = ""

    _name: str = ""
    ft: str = ""
    album: str = ""
    catalognum: str = ""
    remix: Optional[Remix] = None

    digi_only: bool = False
    track_alt: Optional[str] = None

    @classmethod
    def from_json(cls, json: JSONDict, delim: str, label: str) -> "Track":
        try:
            artist = json["inAlbum"]["byArtist"]["name"]
        except KeyError:
            artist = ""
        artist = artist or json.get("byArtist", {}).get("name", "")
        data = {
            "json_item": json,
            "json_artist": artist,
            "track_id": json["@id"],
            "index": json.get("position"),
            "catalognum": json["name_parts"].get("catalognum", ""),
        }
        return cls(**cls.parse_name(data, json["name_parts"]["clean"], delim, label))

    @staticmethod
    def clean_digi_name(name: str) -> Tuple[str, bool]:
        """Clean the track title from digi-only artifacts.

        Return the clean name, and whether this track is digi-only.
        """
        clean_name = DIGI_ONLY_PATTERN.sub("", name)
        return clean_name, clean_name != name

    @staticmethod
    def find_featuring(data: JSONDict) -> JSONDict:
        """Find featuring artist in the track name.

        If the found artist is contained within the remixer, do not do anything.
        If the found artist is among the main artists, remove it from the name but
        do not consider it as a featuring artist.
        Otherwise, strip brackets and spaces and save it in the 'ft' field.
        """
        for _field in "_name", "json_artist":
            m = PATTERNS["ft"].search(data[_field])
            if m:
                ft = m.groups()[-1].strip()
                if ft not in data.get("remixer", ""):
                    data[_field] = data[_field].replace(m.group().rstrip(), "")
                    if ft not in data["json_artist"]:
                        data["ft"] = m.group().strip(" ([])")
                    break
        return data

    @staticmethod
    def parse_name(data: JSONDict, name: str, delim: str, label: str) -> JSONDict:
        name = name.replace(f" {delim} ", " - ")

        # remove label from the end of the track name
        # see https://gutterfunkuk.bandcamp.com/album/gutterfunk-all-subject-to-vibes-various-artists-lp  # noqa
        if name.endswith(label):
            name = name.replace(label, "").strip(" -")

        json_artist, artist_digi_only = Track.clean_digi_name(data["json_artist"])
        name, name_digi_only = Track.clean_digi_name(name)
        data["digi_only"] = name_digi_only or artist_digi_only

        data["json_artist"] = Helpers.clean_name(json_artist) if json_artist else ""
        name = Helpers.clean_name(name).strip().lstrip("-")

        m = PATTERNS["track_alt"].search(name)
        if m:
            data["track_alt"] = m.group(1).replace(".", "").upper()
            name = name.replace(m.group(), "")

        if not data.get("catalognum"):
            # check whether track name contains the catalog number within parens
            # or square brackets
            # see https://objection999x.bandcamp.com/album/eruption-va-obj012
            m = CATNUM_PAT["delimited"].search(name)
            if m:
                data["catalognum"] = m.group(1)
                name = name.replace(m.group(), "").strip()
        name = re.sub(rf"^0*{data.get('index', 0)}(?!\W\d)\W+", "", name)

        remix = Remix.from_name(name)
        if remix:
            data.update(remix=remix)
            name = name.replace(remix.delimited, "").rstrip()

        for m in ELP_ALBUM_PAT.finditer(name):
            data["album"] = m.group(1).replace('"', "")
            name = name.replace(m.group(), "")

        data["_name"] = name
        data = Track.find_featuring(data)
        return data

    @cached_property
    def duration(self) -> Optional[int]:
        try:
            h, m, s = map(int, re.findall(r"\d+", self.json_item["duration"]))
        except KeyError:
            return None
        else:
            return h * 3600 + m * 60 + s

    @cached_property
    def lyrics(self) -> str:
        try:
            text: str = self.json_item["recordingOf"]["lyrics"]["text"]
        except KeyError:
            return ""
        else:
            return text.replace("\r", "")

    @cached_property
    def name(self) -> str:
        name = self._name
        if self.json_artist and " - " not in name:
            name = f"{self.json_artist} - {name}"
        return name.strip()

    @cached_property
    def main_title(self) -> str:
        """Split the track name, deduce the title and return it.
        The extra complexity here is to ensure that it does not cut off a title
        that ends with ' - -', like in '(DJ) NICK JERSEY - 202memo - - -'.
        """
        parts = re.split(r" - (?![^\[(]+[])])", self.name)
        if len(parts) == 1:
            parts = self.name.split(" - ")
        main_title = parts[-1]
        for idx, maybe in enumerate(reversed(parts)):
            if not maybe.strip(" -"):
                main_title = " - ".join(parts[-idx - 2 :])
                break
        return main_title

    @cached_property
    def title(self) -> str:
        """Return the main title with the full remixer part appended to it."""
        if self.remix:
            return f"{self.main_title} {self.remix.delimited}"
        return self.main_title

    @cached_property
    def artist(self) -> str:
        """Take the name, remove the title, ensure it does not duplicate any remixers
        and return the resulting artist.
        """
        artist = self.name[: self.name.rfind(self.main_title)].strip(", -")
        artist = Remix.PATTERN.sub("", artist)
        if self.remix:
            artist = artist.replace(self.remix.remixer, "").strip(" ,")
        return artist.strip(" -")

    @property
    def artists(self) -> List[str]:
        return Helpers.split_artists([self.artist])

    @property
    def info(self) -> JSONDict:
        return {
            "index": self.index,
            "medium_index": self.index,
            "medium": None,
            "track_id": self.track_id,
            "artist": self.artist + (f" {self.ft}" if self.ft else ""),
            "title": self.title,
            "length": self.duration,
            "track_alt": self.track_alt,
            "lyrics": self.lyrics,
            "catalognum": self.catalognum or None,
        }
